-- This file is part of Hypha.
--
-- Hypha is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Hypha is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Hypha.  If not, see <http://www.gnu.org/licenses/>.
--
-- Copyright 2013-2014 Luca Padovani

module Main (main) where

import Process
import Render
import Linearity.Use
import Linearity.Type
import Linearity.UseCombination
import Linearity.TypeCombination
import Linearity.Process

import qualified Parser
import qualified Lexer
import qualified Linearity.Generator as Generator
import qualified Relation
import Control.Monad
import System.Console.GetOpt
import System.IO
import System.Exit
import System.Environment
import Data.Set.Unicode
import qualified Data.Set as S
import qualified Data.Map as M
import qualified Data.List as L
import qualified Linearity.UseSolver as UseSolver
import qualified Linearity.TypeSolver as TypeSolver
import qualified Linearity.Partitioner as Partitioner
import qualified Linearity.Combination as Combination
import Data.Time

import Debug.Trace

versionInfo :: String
versionInfo =
    "Hypha 0.3 - Type Reconstruction for the Linear π-Calculus\n"
    ++ "Copyright © 2013-2014 Luca Padovani\n"
    ++ "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.\n"
    ++ "This is free software: you are free to change and redistribute it.\n"
    ++ "There is NO WARRANTY, to the extent permitted by law."

showPRED :: (Show a) => String -> a -> String
showPRED pred x = pred ++ " " ++ show x

showREL :: (Show a, Show b) => String -> (a, b) -> String
showREL rel (x, y) = "  " ++ show x ++ " " ++ rel ++ " " ++ show y

showSDV :: (TypeV, (Type, Type)) -> String
showSDV (tvar, (t, s)) = "  " ++ show (TVar tvar :: Type) ++ " = " ++ show t ++ " - " ++ show s

showC :: (TypeV, S.Set TypeC) -> String
showC (tvar, tcset) = "  " ++ show (TVar tvar :: Type) ++ " = " ++ showSet tcset
    where
      showSet :: Show a => S.Set a -> String
      showSet s = L.intercalate " = " (L.map show $ S.toList s)

putSectionTitle :: String -> IO ()
putSectionTitle s = do putStrLn $ "┌─" ++ L.replicate (length s) '─' ++ "─╖"
                       putStrLn $ "│ " ++ s ++ " ║"
                       putStrLn $ "╘═" ++ L.replicate (length s) '═' ++ "═╝"

printUseEnvironment :: Use -> UseEnvironment -> IO ()
printUseEnvironment u env =
    do let us = map (show . UVar . fst) $ M.toList $ M.filter (== u) env
       if us /= [] then putStrLn $ "  " ++ show u ++ (concat $ map (" = " ++) us)
       else return ()

printTime :: String -> UTCTime -> IO ()
printTime msg t0 =
    do t1 <- getCurrentTime
       putStrLn $ "⌚ [" ++ msg ++ ": " ++ show (diffUTCTime t1 t0) ++ "]"

reconstruction :: [Flag] -> UntypedProcess -> IO (Process Type, M.Map Name Type)
reconstruction args p =
    do timeGenerateStart <- getCurrentTime
       let (q, env, (eqm0, stm0, cot0, com0, eqU)) = Generator.generate (OddChannels `elem` args) p
       when (Verbose `elem` args)
            (do putSectionTitle "PROCESS"
                putStrLn $ Render.renderProcess show q
                putSectionTitle "UNSOLVED TYPE ENVIRONMENT"
                forM_ (M.toAscList env) (\(u, t) -> putStrLn $ "  " ++ u ++ " : " ++ show t)
                putSectionTitle "TYPE CONSTRAINTS"
                forM_ (map (\(tvar, t) -> (TVar tvar :: Type, t)) $ M.toList eqm0) (putStrLn . showREL "=")
                forM_ (map (\(tvar, t) -> (TVar tvar :: Type, t)) $ M.toList $ M.difference stm0 eqm0) (putStrLn . showREL "~")
                forM_ (map (\(tvar, tc) -> (TVar tvar :: Type, tc)) $ M.toList cot0) (putStrLn . showREL "=")
                putSectionTitle "ALL TYPE COMBINATIONS"
                forM_ (M.toList com0) (putStrLn . showC)
                putSectionTitle "USE CONSTRAINTS"
                forM_ (S.toList eqU) (putStrLn . showREL "=")
                when (Timing `elem` args)
                     (printTime "CONSTRAINT GENERATION" timeGenerateStart)
            )

       timeUseSolutionStart <- getCurrentTime
       let ucS = if Partition `elem` args then Partitioner.partitionUseConstraints eqU else [eqU]
       let uenv0 = M.unions $ map (UseSolver.findUseSolution (MinUses `elem` args)) ucS
       let uvarset = (S.unions $ map Linearity.Type.fuv $ M.elems eqm0) ∪
                     (S.unions $ map Linearity.Type.fuv $ M.elems stm0)
       let uenv = M.union uenv0 $ M.fromList [(uvar, 0) | uvar <- S.toList uvarset ]

       when (Verbose `elem` args)
            (do putSectionTitle "USE SOLUTION"
                putStrLn $ "  [found " ++ show (length ucS) ++ " partition(s) of use constraints]"
                printUseEnvironment 0 uenv
                printUseEnvironment 1 uenv
                printUseEnvironment ω uenv
                when (Timing `elem` args)
                     (printTime "USE SOLUTION" timeUseSolutionStart)
            )

       let eqm = M.map (substUT uenv) eqm0
       let stm = M.map (substUT uenv) stm0
       -- lazyness alert: if there is no solution for use constraints but there are
       -- empty combinations in front of those where the unsolvable use variables occur
       -- then the program will not signal unsatisfiability and will output a (wrong) result
       -- the following line will FORCE evaluation of all the use constraints even when
       -- hypha has been invoked without the -v option
       when (minimum (M.elems uenv) >= Zero) (return ())
       let com = M.map (Linearity.TypeCombination.substUC uenv . S.findMin) com0
       timeTypeSolutionStart <- getCurrentTime
       let defm = TypeSolver.definitions eqm stm com
       when (Verbose `elem` args)
            (do putSectionTitle "TYPE SOLUTION"
                forM_ (map (\(tvar, t) -> (TVar tvar :: Type, t)) $ M.toList defm) (putStrLn . showREL "↦")
                when (Timing `elem` args)
                     (do printTime "TYPE SOLUTION" timeTypeSolutionStart
                         printTime "RECONSTRUCTION" timeGenerateStart)
            )

       let f = substTT defm . substUT uenv
       return (mapProcess f q, M.map f env)

main :: IO ()
main = do
  (args, file) <- getArgs >>= parse
  source <- if file == "-" then getContents else readFile file
  let p = Parser.process $ Lexer.alexScanTokens source

  timeProcessStart <- getCurrentTime
  (q, env) <- reconstruction args p

  putSectionTitle "SOLVED TYPE ENVIRONMENT"
  forM_ (M.toAscList env) (\(u, t) -> putStrLn $ "  " ++ u ++ " : " ++ show t)
  timeProcessEnd <- getCurrentTime
  when (Timing `elem` args)
       (hPutStrLn stderr $ show (diffUTCTime timeProcessEnd timeProcessStart) ++ " for processing " ++ file)

data Flag = Partition   --    --partition
          | OddChannels -- -o --odd-channels
          | MinUses     -- -m --minimum-uses
          | Timing      --    --timing
          | Verbose     -- -v --verbose
          | Version     -- -V --version
          | Help        --    --help
            deriving (Eq,Ord,Enum,Show,Bounded)
 
flags =
   [Option ""  ["partition"]      (NoArg Partition)   "Partition use constraints",
    Option "o" ["odd-channels"]   (NoArg OddChannels) "Allow different input/output uses for restricted channels",
    Option "m" ["minimum-uses"]   (NoArg MinUses)     "Minimize the number of ωs in uses",
    Option ""  ["timing"]         (NoArg Timing)      "Show timing",
    Option "v" ["verbose"]        (NoArg Verbose)     "Print use and type constraints",
    Option "V" ["version"]        (NoArg Version)     "Print version information",
    Option "h" ["help"]           (NoArg Help)        "Print this help message"
   ]
 
parse argv =
    case getOpt Permute flags argv of
      (args, files, []) -> do
        when (Version `elem` args)
             (do hPutStrLn stderr versionInfo
                 exitWith ExitSuccess)
        when (null files || L.length files > 1 || Help `elem` args)
                 (do hPutStrLn stderr (usageInfo header flags)
                     exitWith ExitSuccess)
        return (args, head files)
 
      (_, _, errs) -> do
        hPutStrLn stderr (concat errs ++ usageInfo header flags)
        exitWith (ExitFailure 1)
 
    where
      header = "Usage: hypha [options] FILE"
